-- ========= --
-- UTILITIES --
-- ========= --

local iPracticalNumCivs = (GameDefines.MAX_MAJOR_CIVS - 1)

function JFD_IsCivilisationActive(civilizationID)
	for iSlot = 0, iPracticalNumCivs, 1 do
		local slotStatus = PreGame.GetSlotStatus(iSlot)
		if (slotStatus == SlotStatus["SS_TAKEN"] or slotStatus == SlotStatus["SS_COMPUTER"]) then
			if PreGame.GetCivilization(iSlot) == civilizationID then
				return true
			end
		end
	end
	return false
end

function JFDGame_IsAAActive()
	for _, mod in pairs(Modding.GetActivatedMods()) do
		if mod.ID == "432bc547-eb05-4189-9e46-232dbde8f09f" then
			return true
		end
	end
	return false
end
local isAAActive = JFDGame_IsAAActive()

function Neirai_GetNearestCity(pPlayer, pPlot)
	local distance = 9999
	local cNearestCity = nil
	for cCity in pPlayer:Cities() do
		local pCityPlot = cCity:Plot()
		local between = Map.PlotDistance(pPlot:GetX(), pPlot:GetY(), pCityPlot:GetX(), pCityPlot:GetY())
		if between < distance then
			distance = between
			cNearestCity = cCity
		end
	end
	return cNearestCity
end

function JFD_GetRandom(lower, upper)
    return Game.Rand((upper + 1) - lower, "") + lower
end

-- ======= --
-- DEFINES --
-- ======= --

include("AdditionalAchievementsUtility.lua")
include("FLuaVector.lua")
include("IconSupport.lua")
include("InstanceManager.lua")

local iCiv = GameInfoTypes.CIVILIZATION_THP_SOKOTO
local bIsActive = JFD_IsCivilisationActive(iCiv)

-- =========================== --
-- UA: TRACK CITADEL POSITIONS --
-- =========================== --

local impCitadel = GameInfoTypes.IMPROVEMENT_CITADEL
local tCitadels = {}

function TabulateCitadels()
	for i = 0, Map:GetNumPlots() - 1, 1 do
		local pPlot = Map.GetPlotByIndex(i)
		if pPlot:GetImprovementType() == impCitadel then
			tCitadels[pPlot] = true
		end
	end
end

function UpdateCitadelsOnNewBuild(playerID, iX, iY, impType)
	local pPlot = Map.GetPlot(iX, iY)
	if impType == impCitadel then
		tCitadels[pPlot] = true
	elseif tCitadels[pPlot] then
		tCitadels[pPlot] = false
	end
end

if bIsActive then
	Events.SequenceGameInitComplete.Add(TabulateCitadels)
	GameEvents.BuildFinished.Add(UpdateCitadelsOnNewBuild)
end

-- ======================================== --
-- UA: CITADELS BOOST ADJACENT IMPROVEMENTS --
-- ======================================== --

local iNumDirections = DirectionTypes.NUM_DIRECTION_TYPES - 1

local yFood = GameInfoTypes.YIELD_FOOD
local yProd = GameInfoTypes.YIELD_PRODUCTION
local yGold = GameInfoTypes.YIELD_GOLD
local ySci = GameInfoTypes.YIELD_SCIENCE
local yCult = GameInfoTypes.YIELD_CULTURE
local yFaith = GameInfoTypes.YIELD_FAITH

local tYieldTypes = {}
local iYTValue = 1
for row in DB.Query("SELECT ID FROM Yields") do
	tYieldTypes[iYTValue] = row.ID
	iYTValue = iYTValue + 1
end

function AddToYield(pPlayer, pPlot, yType)
	if yType == yFood then
		local cCity = Neirai_GetNearestCity(pPlayer, pPlot)
		cCity:ChangeFood(1)
	elseif yType == yProd then
		local cCity = Neirai_GetNearestCity(pPlayer, pPlot)
		cCity:ChangeProduction(1)
	elseif yType == yGold then
		pPlayer:ChangeGold(1)
	elseif yType == ySci then
		local tTeamTech = Teams[pPlayer:GetTeam()]:GetTeamTechs()
		local techCurrent = pPlayer:GetCurrentResearch()
		tTeamTech:ChangeResearchProgress(techCurrent, 1, pPlayer)
	elseif yType == yCult then
		pPlayer:ChangeJONSCulture(1)
	else
		--[[    Ultimately, if there's a way to make this fully modular, I haven't figured
		--	it out. So, what I've decided to do is just have this default to faith if it gets
		--	a new type of yield added by some other mod.]]
		pPlayer:ChangeFaith(1)
	end
end

function CitadelYieldBoost(playerID)
	local pPlayer = Players[playerID]
	if pPlayer:GetCivilizationType() == iCiv then
		for kPlot, v in pairs(tCitadels) do
			if (kPlot:GetOwner() == playerID) and (not kPlot:IsImprovementPillaged()) then
				for iDirection = 0, iNumDirections, 1 do
					local pAdjPlot = Map.PlotDirection(kPlot:GetX(), kPlot:GetY(), iDirection)
					if (pAdjPlot:GetOwner() == playerID) and (pAdjPlot:GetImprovementType() > -1) and (not pAdjPlot:IsCity()) then
						for k, vYield in pairs(tYieldTypes) do
							local iBaseYield = pAdjPlot:GetYield(vYield)
							if iBaseYield > 0 then
								AddToYield(pPlayer, pAdjPlot, vYield)
							end
						end
					end
				end
			end
		end
	end
end

if bIsActive then
	GameEvents.PlayerDoTurn.Add(CitadelYieldBoost)
end

-- ================================= --
-- UA: GG POINTS FROM FAITH PURCHASE --
-- ================================= --

function SokotoFaithBoughtBuilding(playerID, cityID, bldgType, bGold, bFaith)
	if bFaith then
		local pPlayer = Players[playerID]
		if pPlayer:GetCivilizationType() == iCiv then
			local iCost = GameInfo.Buildings[bldgType].FaithCost
			local iPointGain = math.floor(iCost / 20)
			pPlayer:ChangeCombatExperience(iPointGain)
			if pPlayer:IsHuman() then
				Events.GameplayAlertMessage("Your [ICON_PEACE] Faith purchase has generated " .. iPointGain .. " [ICON_GREAT_PEOPLE] Great General points!")
			end
		end
	end
end

function SokotoFaithBoughtUnit(playerID, cityID, unitID, bGold, bFaith)
	if bFaith then
		local pPlayer = Players[playerID]
		if pPlayer:GetCivilizationType() == iCiv then
			local uNewUnit = pPlayer:GetUnitByID(unitID)
			local pPlot = pPlayer:GetCityByID(cityID):Plot()
			for i = 0, pPlot:GetNumUnits() - 1, 1 do
				local uPlotUnit = pPlot:GetLayerUnit(i)
				if uPlotUnit == uNewUnit then
					if uNewUnit:IsGreatPerson() then
						-- b/c Great People don't have faith costs per se,
						-- I just defaulted to 1000 which iirc is where they start
						pPlayer:ChangeCombatExperience(50)
						if pPlayer:IsHuman() then
							Events.GameplayAlertMessage("Your [ICON_PEACE] Faith purchase has generated 50 [ICON_GREAT_PEOPLE] Great General points!")
						end
					else
						local uType = uNewUnit:GetUnitType()
						local iCost = GameInfo.Units[uType].FaithCost
						local iPointGain = math.floor(iCost / 20)
						pPlayer:ChangeCombatExperience(iPointGain)
						if pPlayer:IsHuman() then
							Events.GameplayAlertMessage("Your [ICON_PEACE] Faith purchase has generated " .. iPointGain .. " [ICON_GREAT_PEOPLE] Great General points!")
						end
					end
				end
			end
		end
	end
end

if bIsActive then
	GameEvents.CityConstructed.Add(SokotoFaithBoughtBuilding)
	GameEvents.CityTrained.Add(SokotoFaithBoughtUnit)
end

-- =============== --
-- UU1: YAN LIFIDA --
-- =============== --

-- N/A; this unit's abilities are all done through SQL

-- ====================================== --
-- UU2: MAY BE EXPENDED ON FORTS/CITADELS --
-- ====================================== --

local impFort = GameInfoTypes.IMPROVEMENT_FORT
local uMarabout = GameInfoTypes.UNIT_THP_MARABOUT
local ucMissionary = GameInfoTypes.UNITCLASS_MISSIONARY
local pActivePlayer = Players[Game.GetActivePlayer()]

function IsButtonPossible(uUnit)
	if uUnit:GetUnitType() == uMarabout then
		local pPlot = uUnit:GetPlot()
		if pPlot then
			local impOnPlot = pPlot:GetImprovementType()
			if (impOnPlot == impFort) or (impOnPlot == impCitadel) then
				if uUnit:GetOwner() == pPlot:GetOwner() then
					return true
				end
			end
		end
	end
	return false
end

-- This function grants the requisite faith and science
function EnactFortFaith(pPlayer, uUnit)
	local impType = uUnit:GetPlot():GetImprovementType()
	local iEraMultiplier = pPlayer:GetCurrentEra() + 1
	local tTeamTech = Teams[pPlayer:GetTeam()]:GetTeamTechs()
	local techCurrent = pPlayer:GetCurrentResearch()
	local iFaithQty = 0
	if impType == impCitadel then
		iFaithQty = 10 * iEraMultiplier
		pPlayer:ChangeFaith(iFaithQty)
		tTeamTech:ChangeResearchProgressPercent(techCurrent, 20, pPlayer)
	else
		iFaithQty = 5 * iEraMultiplier
		pPlayer:ChangeFaith(iFaithQty)
		tTeamTech:ChangeResearchProgressPercent(techCurrent, 10, pPlayer)
	end
	if pPlayer:IsHuman() then
		local sMessage = "text error"
		if impType == impCitadel then
			sMessage = "+" .. iFaithQty .. " [ICON_PEACE] and[NEWLINE]+20% [ICON_RESEARCH] of current tech"
		else
			sMessage = "+" .. iFaithQty .. " [ICON_PEACE] and[NEWLINE]+10% [ICON_RESEARCH] of current tech"
		end
		local iX = uUnit:GetX()
		local iY = uUnit:GetY()
		Events.AddPopupTextEvent(HexToWorld(ToHexFromGrid(Vector2(iX, iY))), sMessage, 0)

		if isAAActive then
			SokotoAchievementCheck(pPlayer, iX, iY)
		end
	end
	uUnit:Kill()
end

function Marabout_UnitActionButton()
	local uUnit = UI.GetHeadSelectedUnit();
	EnactFortFaith(pActivePlayer, uUnit)
end

function Marabout_SerialEventUnitInfoDirty()
	local uUnit = UI.GetHeadSelectedUnit();
	if (not uUnit) then return end
	if IsButtonPossible(uUnit) then
		Controls.UnitActionButton:SetHide(false)
	else
		Controls.UnitActionButton:SetHide(true)
	end
	
	local buildCityButtonActive = uUnit:IsFound();
				
	local primaryStack = ContextPtr:LookUpControl("/InGame/WorldView/UnitPanel/PrimaryStack")
	local primaryStretchy = ContextPtr:LookUpControl("/InGame/WorldView/UnitPanel/PrimaryStretchy")
	primaryStack:CalculateSize();
	primaryStack:ReprocessAnchoring();

	local stackSize = primaryStack:GetSize();
	local stretchySize = primaryStretchy:GetSize();
	local buildCityButtonSize = 0
	if buildCityButtonActive then
		if OptionsManager.GetSmallUIAssets() and not UI.IsTouchScreenEnabled() then
			buildCityButtonSize = 36;
		else
			buildCityButtonSize = 60;
		end
	end
	primaryStretchy:SetSizeVal( stretchySize.x, stackSize.y + buildCityButtonSize + 348 );
end

local function Marabout_UpdateUnitInfoPanel()
	if (not OptionsManager.GetSmallUIAssets()) then
		Controls.UnitActionButton:SetSizeVal(50,50)
		Controls.UnitActionIcon:SetSizeVal(64,64)
		Controls.UnitActionIcon:SetTexture("UnitActions360.dds")
	else
		Controls.UnitActionButton:SetSizeVal(36,36)
		Controls.UnitActionIcon:SetSizeVal(45,45)
		Controls.UnitActionIcon:SetTexture("UnitActions360.dds")
	end
	Controls.UnitActionIcon:LocalizeAndSetToolTip("TXT_KEY_THP_MARABOUT_BUTTON_TOOLTIP")
	Controls.UnitActionButton:ChangeParent(ContextPtr:LookUpControl("/InGame/WorldView/UnitPanel/PrimaryStack"))
end

function Marabout_DoTurn(playerID)
	local pPlayer = Players[playerID]
	if pPlayer:HasUnitOfClassType(ucMissionary) then
		if not pPlayer:IsHuman() then
			-- if player is AI, determine if it wants to use the ability
			--[[ A non-Sokoto AI will never choose this. I don't plan to fix this, since
			     that'd require iterating through units whenever an AI player owns a
			     missionary, which would slow down AI turns significantly. --]]
			if pPlayer:GetCivilizationType() ~= iCiv then return end
	                if not pPlayer:GetCapitalCity() then return end
			for uUnit in pPlayer:Units() do
				if IsButtonPossible(uUnit) then
					local impOnPlot = uUnit:GetPlot():GetImprovementType()
					if impOnPlot == impCitadel then
						EnactFortFaith(pPlayer, uUnit)
					else
						if JFD_GetRandom(1, 3) > 1 then
							EnactFortFaith(pPlayer, uUnit)
						end
					end
				end
			end
		end
	end
end

local function Initialize()	
	Events.LoadScreenClose.Add(Marabout_UpdateUnitInfoPanel);
	Events.SerialEventUnitInfoDirty.Add(Marabout_SerialEventUnitInfoDirty);
end

if bIsActive then
	Controls.UnitActionButton:RegisterCallback(Mouse.eLClick, Marabout_UnitActionButton);
	IconHookup(23, 45, "UNIT_ACTION_ATLAS", Controls.UnitActionIcon)
	GameEvents.PlayerDoTurn.Add(Marabout_DoTurn)
	Initialize();
end

-- ========== --
-- AA SUPPORT --
-- ========== --

if isAAActive then
	-- the IsHuman() check is done in EnactFortFaith
	-- so I skip it here to avoid redundancy
	function SokotoAchievementCheck(pPlayer, iX, iY)
		if IsAAUnlocked('AA_THP_SOKOTO_SPECIAL') then return end
		for iDirection = 0, iNumDirections, 1 do
			local pAdjPlot = Map.PlotDirection(iX, iY, iDirection)
			if pAdjPlot:IsCity() and pAdjPlot:GetOwner() ~= pPlayer:GetID() then
				UnlockAA('AA_THP_SOKOTO_SPECIAL');
				return
			end
		end
	end
end
